/*!\file: control_core.cpp
 * \brief: core of the control solution 
 */ 

#include "./solutions.h"
#include "../modules/modules.h"
#include "../EnumDefinitions/EnumDefinitions.h"

DataSet* control_core(Model* model){

	extern int my_rank;

	/*fem model: */
	FemModel* fem_model=NULL;

	/*output: */
	DataSet* results=NULL;
	Result* result=NULL;

	/*Intermediary: */
	DataSet* diagnostic_results=NULL;
	DataSet* gradjcompute_results=NULL;
	DataSet* steadystate_results=NULL;
	Vec     u_g=NULL;
	Vec     t_g=NULL;
	Vec     m_g=NULL;
	double  search_scalar;
	int     control_type;
	double* fit=NULL;
	double* optscal=NULL;
	int     gsize;
	double* maxiter=NULL;
	double* cm_jump=NULL;
	double  eps_cm;
	double  tolx;
	double* param_g=NULL;
	Vec     grad_g=NULL;
	Vec     new_grad_g=NULL;
	Vec     grad_g_old=NULL;
	double* grad_g_double=NULL;
	double  cm_min;
	double  cm_max;
	int     cm_gradient;
	int     nsteps,n,i;
	double* J=NULL;
	OptArgs optargs;
	OptPars optpars;
	Param*  param=NULL;

	/*flags: */
	int analysis_type;
	int sub_analysis_type;
	int     control_steady;
	int verbose=0;
	int converged=0;
	int numberofnodes;

	//initialize results
	results=new DataSet(ResultsEnum);

	/*Process models*/
	ControlInitialization(model);
	fem_model=model->GetActiveFormulation();

	/*Recover parameters used throughout the solution:*/
	model->FindParam(&nsteps,NStepsEnum);
	model->FindParam(&control_type,ControlTypeEnum);
	model->FindParam(&fit,NULL,NULL,FitEnum);
	model->FindParam(&optscal,NULL,NULL,OptScalEnum);
	model->FindParam(&maxiter,NULL,NULL,MaxIterEnum);
	model->FindParam(&cm_jump,NULL,NULL,CmJumpEnum);
	model->FindParam(&eps_cm,EpsCmEnum);
	model->FindParam(&tolx,TolXEnum);
	model->FindParam(&cm_min,CmMinEnum);
	model->FindParam(&cm_max,CmMaxEnum);
	model->FindParam(&cm_gradient,CmGradientEnum);
	model->FindParam(&param_g,NULL,NULL,ControlParameterEnum);
	model->FindParam(&analysis_type,AnalysisTypeEnum);
	model->FindParam(&sub_analysis_type,SubAnalysisTypeEnum);
	model->FindParam(&numberofnodes,NumberOfNodesEnum);
	model->FindParam(&control_steady,ControlSteadyEnum);
	gsize=fem_model->nodes->NumberOfDofs();

	/*Initialize misfit: */
	J=(double*)xmalloc(nsteps*sizeof(double));

	/*Start looping: */
	for(n=0;n<nsteps;n++){

		_printf_("\n%s%i%s%i\n","   control method step ",n+1,"/",nsteps);
		model->UpdateInputsFromVector(param_g,control_type,VertexEnum);
		model->UpdateInputsFromConstant(fit[n],FitEnum);
		
		/*In case we are running a steady state control method, compute new temperature field using new parameter 
		 * distribution: */
		if (control_steady){
			steadystate_results= steadystate_core(model);
			VecFree(&t_g); steadystate_results->FindResult(&t_g,"t_g");
			delete steadystate_results;
			model->UpdateInputsFromVector(t_g,TemperatureEnum,VertexEnum);
		}
	
		_printf_("%s\n","      computing gradJ...");
		gradjcompute_results= gradjcompute_core(model);
		gradjcompute_results->FindResult(&grad_g,"grad_g");
		delete gradjcompute_results;

		/*Return gradient if asked: */
		if (cm_gradient){
			
			/*Plug results into output dataset: */
			result=new Result(results->Size()+1,0,1,"grad_g",grad_g);
			results->AddObject(result);
	
			result=new Result(results->Size()+1,0,1,"analysis_type",EnumAsString(analysis_type));
			results->AddObject(result);
	
			/*Free ressources: */
			xfree((void**)&control_type);
			xfree((void**)&fit);
			xfree((void**)&optscal);
			xfree((void**)&maxiter);
			xfree((void**)&cm_jump);
			xfree((void**)&grad_g_double);
			xfree((void**)&param_g);
			VecFree(&u_g);
			VecFree(&t_g);
			VecFree(&m_g);
			xfree((void**)&J);
			return results;
		}

		/*Normalize if last gradient not satisfying (search_scalar==0)*/
		if (n>0 && search_scalar==0){
			_printf_("%s","      orthogonalization...");
			Orthx(&new_grad_g,grad_g,grad_g_old);
			_printf_("%s\n"," done.");
		}
		else{ 
			_printf_("%s","      normalizing directions...");
			Orthx(&new_grad_g,grad_g,NULL);
			_printf_("%s\n"," done.");
		}
		VecFree(&grad_g); VecFree(&grad_g_old); 
		grad_g_old=new_grad_g;
		VecToMPISerial(&grad_g_double,new_grad_g);

		_printf_("%s\n","      optimizing along gradient direction");
		optargs.model=model;
		optargs.param_g=param_g; optargs.grad_g=grad_g_double; optargs.n=n;
		optpars.xmin=0; optpars.xmax=1; optpars.tolerance=tolx; optpars.maxiter=(int)maxiter[n];optpars.cm_jump=cm_jump[n];
		BrentSearch(&search_scalar,J+n,&optpars,&objectivefunctionC,&optargs);

		_printf_("%s","      updating parameter using optimized search scalar...");
		for(i=0;i<numberofnodes;i++)param_g[i]=param_g[i]+search_scalar*optscal[n]*grad_g_double[i];
		_printf_("%s\n"," done.");

		_printf_("%s","      constraining the new distribution...");    
		ControlConstrainx(param_g,numberofnodes,cm_min,cm_max,control_type);
		_printf_("%s\n"," done.");
		
		_printf_("%s%i%s%g\n","      value of misfit J after optimization #",n+1,": ",J[n]);

		/*some freeing:*/
		xfree((void**)&grad_g_double);

		/*Has convergence been reached?*/
		if (!isnan(eps_cm)){
			i=n-2;
			//go through the previous misfits(starting from n-2)
			while(i>=0){
				if (fit[i]==fit[n]){
					//convergence test only if we have the same misfits
					if ((J[i]-J[n])/J[n] <= eps_cm){
						//convergence if convergence criteria fullfilled
						converged=1;
						_printf_("%s%g%s%g\n","      Convergence criterion: dJ/J = ",(J[i]-J[n])/J[n],"<",eps_cm);
					}
					else{
						_printf_("%s%g%s%g\n","      Convergence criterion: dJ/J = ",(J[i]-J[n])/J[n],">",eps_cm);
					}
					break;
				}
				i=i-1;
			}
		}
		//stop if convergence has been reached
		if(converged) break;

		//some temporary saving
		if (((n+1)%5)==0){
			_printf_("%s","      saving temporary results...");
			ControlRestart(model,param_g);
			_printf_("%s\n"," done.");
		}
	}

	/*Write results to disk: */
	_printf_("%s","      preparing final velocity solution");
	/*Launch diagnostic with the last parameter distribution*/
	if (control_steady){
		model->UpdateInputsFromVector(param_g,control_type,VertexEnum);
		steadystate_results=steadystate_core(model);

		//extract u_g ,t_g and m_g from steadystate results, and erase diagnostic_results;
		steadystate_results->FindResult(&u_g,"u_g");
		steadystate_results->FindResult(&m_g,"m_g");
		steadystate_results->FindResult(&t_g,"t_g");
		delete steadystate_results;
	}
	else{
		model->UpdateInputsFromVector(param_g,control_type,VertexEnum);
		diagnostic_results=diagnostic_core(model);

		//extract u_g from diagnostic_results, and erase diagnostic_results;
		diagnostic_results->FindResult(&u_g,"u_g");
		delete diagnostic_results;
	}

	/*Plug results into output dataset: */
	result=new Result(results->Size()+1,0,1,"u_g",u_g);
	results->AddObject(result);
	result=new Result(results->Size()+1,0,1,"param_g",param_g,numberofnodes);
	results->AddObject(result);
	result=new Result(results->Size()+1,0,1,"J",J,nsteps);
	results->AddObject(result);
	if (control_steady){
		result=new Result(results->Size()+1,0,1,"t_g",t_g);
		results->AddObject(result);
		result=new Result(results->Size()+1,0,1,"m_g",m_g);
		results->AddObject(result);
	}
	
	/*Add analysis_type and control_type to results: */
	result=new Result(results->Size()+1,0,1,"analysis_type",EnumAsString(analysis_type));
	results->AddObject(result);
	result=new Result(results->Size()+1,0,1,"control_type",EnumAsString(control_type));
	results->AddObject(result);


	/*Free ressources: */
	xfree((void**)&control_type);
	xfree((void**)&fit);
	xfree((void**)&optscal);
	xfree((void**)&maxiter);
	xfree((void**)&cm_jump);
	VecFree(&new_grad_g); //do not VecFree grad_g and grad_g_old, they point to new_grad_g
	xfree((void**)&grad_g_double);
	xfree((void**)&param_g);
	VecFree(&u_g);
	VecFree(&t_g);
	VecFree(&m_g);
	xfree((void**)&J);

	//return: 
	return results;
}
